table.nydok-requirements {
    margin-top: 1rem;
    margin-bottom: 1rem;
    width: 90%;
    margin-left: auto;
    margin-right: auto;
}

table.nydok-requirements td.req-id {
    background-color: rgba(240, 158, 120, 0.4);
    background-color: #FD8143;
    color: #fff;
    font-weight: bold;
    width: 1%;
}

table.nydok-requirements td.req-refs {
    background-color: rgba(240, 158, 120, 0.7);
    background-color: #FF9864;
    color: #fff;
    width: 1%;
}

table.nydok-requirements td.req-desc {
    background-color: rgba(240, 158, 120, 0.15);
    width: 98%;
}