class DuplicateRequirementException(Exception): ...


class FailedTestCaseException(Exception): ...


class MissingTestCaseException(Exception): ...


class RiskPriorityExceedsThresholdException(Exception): ...
